/* $Id: workout_s610.c,v 1.1 2004/11/23 18:55:21 dave Exp $ */

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "s710.h"
#include "config.h"

#define  MAKEBUF(a,b) \
if ( (w->a = calloc(w->samples,sizeof(b))) == NULL ) { \
  fprintf(stderr,"%s: calloc(%d,%ld): %s\n", \
          #a,w->samples,(long)sizeof(b),strerror(errno)); \
  ok = 0; \
}
                     

/* don't call this unless you're sure it's going to be ok. */

workout_t *
extract_workout_s610 ( unsigned char *buf, S710_Filter filter )
{
  int              i;
  int              offset;
  int              s;
  int              x;
  int              recsiz;
  workout_t       *w;
  lap_data_t      *l;
  S710_Distance    prev_lap_dist;
  S710_Altitude    prev_lap_ascent;
  int              ok;
  int              ax;

  if ( (w = calloc(1,sizeof(workout_t))) == NULL ) {
    fprintf(stderr,"extract_workout: calloc(%ld): %s\n",
	    (long)sizeof(workout_t),strerror(errno));
    return NULL;
  }

  /* define the type of the HRM */

  w->type = S710_HRM_S610;

  /* number of bytes in the buffer (including these two) */

  w->bytes = buf[0] + (buf[1]<<8);

  w->exercise_number = buf[2];
  if ( w->exercise_number > 0 && w->exercise_number <= 5 ) {
    extract_label(&buf[3],&w->exercise_label,7);
  } else {
    strcpy(w->exercise_label,"<empty>");
  }
  
  /* date of workout */

  w->date.tm_sec   = BCD(buf[10]);
  w->date.tm_min   = BCD(buf[11]);
  w->date.tm_hour  = BCD(buf[12] & 0x7f);

  /* PATCH for AM/PM mode detection from Berend Ozceri */

  w->ampm = S710_AM_PM_MODE_UNSET;
  if ( buf[13] & 0x80 ) w->ampm |= S710_AM_PM_MODE_SET;
  if ( buf[12] & 0x80 ) w->ampm |= S710_AM_PM_MODE_PM;

  w->date.tm_hour += (buf[13] & 0x80) ?                     /* am/pm mode?   */
    ((buf[12] & 0x80) ? ((w->date.tm_hour < 12) ? 12 : 0) : /* yes, pm set   */
     ((w->date.tm_hour >= 12) ? -12 : 0)) :                 /* yes, pm unset */
    0;                                                      /* no            */

  w->date.tm_mon   = LNIB(buf[15]) - 1;
  w->date.tm_mday  = BCD(buf[13] & 0x7f);
  w->date.tm_year  = 100 + BCD(buf[14]);
  w->date.tm_isdst = -1; /* Daylight savings time not known yet? */
  w->unixtime = mktime(&w->date);

  /* duration of workout */

  w->duration.tenths       = UNIB(buf[15]);
  w->duration.seconds      = BCD(buf[16]);
  w->duration.minutes      = BCD(buf[17]);
  w->duration.hours        = BCD(buf[18]);

  /* avg, max HR */
  
  w->avg_hr                = buf[19];
  w->max_hr                = buf[20];

  /* number of laps */

  w->laps                  = BCD(buf[21]);

  /* user ID */

  w->user_id               = BCD(buf[24]);

  /* units */

  if ( buf[25] & 0x02 ) { /* english */

    w->units.system      = S710_UNITS_ENGLISH;
    w->units.altitude    = S710_ALTITUDE_FT;
    w->units.speed       = S710_SPEED_MPH;
    w->units.distance    = S710_DISTANCE_MI;
    w->units.temperature = S710_TEMPERATURE_F;
    ax                   = 5;

  } else {         /* metric */

    w->units.system      = S710_UNITS_METRIC;
    w->units.altitude    = S710_ALTITUDE_M;
    w->units.speed       = S710_SPEED_KPH;
    w->units.distance    = S710_DISTANCE_KM;
    w->units.temperature = S710_TEMPERATURE_C;
    ax                   = 1;
  }

  /* recording mode and interval */

  w->mode = 0;
  
  switch ( buf[26] ) {
  case 0:  w->recording_interval = 5;  break;
  case 1:  w->recording_interval = 15; break;
  case 2:  w->recording_interval = 60; break;
  default: w->recording_interval = 0;  break;
  }

  /* HR limits */

  w->hr_limit[0].lower     = buf[28];
  w->hr_limit[0].upper     = buf[29];
  w->hr_limit[1].lower     = buf[30];
  w->hr_limit[1].upper     = buf[31];
  w->hr_limit[2].lower     = buf[32];
  w->hr_limit[2].upper     = buf[33];

  /* time below, within, above hr limits 1 */
  
  w->hr_zone[0][0].tenths  = 0;
  w->hr_zone[0][0].seconds = BCD(buf[37]);  /* Below zone 1, seconds */
  w->hr_zone[0][0].minutes = BCD(buf[38]);  /* Below zone 1, minutes */
  w->hr_zone[0][0].hours   = BCD(buf[39]);  /* Below zone 1, hours   */
  w->hr_zone[0][1].seconds = BCD(buf[40]);  /* Within zone 1, seconds */
  w->hr_zone[0][1].minutes = BCD(buf[41]);  /* Within zone 1, minutes */
  w->hr_zone[0][1].hours   = BCD(buf[42]);  /* Within zone 1, hours   */
  w->hr_zone[0][2].seconds = BCD(buf[43]);  /* Above zone 1, seconds */
  w->hr_zone[0][2].minutes = BCD(buf[44]);  /* Above zone 1, minutes */
  w->hr_zone[0][2].hours   = BCD(buf[45]);  /* Above zone 1, hours   */

  /* time below, within, above hr limits 2 */

  w->hr_zone[1][0].tenths  = 0;
  w->hr_zone[1][0].seconds = BCD(buf[46]);  /* Below zone 2, seconds */
  w->hr_zone[1][0].minutes = BCD(buf[47]);  /* Below zone 2, minutes */
  w->hr_zone[1][0].hours   = BCD(buf[48]);  /* Below zone 2, hours   */
  w->hr_zone[1][1].seconds = BCD(buf[49]);  /* Within zone 2, seconds */
  w->hr_zone[1][1].minutes = BCD(buf[50]);  /* Within zone 2, minutes */
  w->hr_zone[1][1].hours   = BCD(buf[51]);  /* Within zone 2, hours   */
  w->hr_zone[1][2].seconds = BCD(buf[52]);  /* Above zone 2, seconds */
  w->hr_zone[1][2].minutes = BCD(buf[53]);  /* Above zone 2, minutes */
  w->hr_zone[1][2].hours   = BCD(buf[54]);  /* Above zone 2, hours   */

  /* time below, within, above hr limits 3 */

  w->hr_zone[2][0].tenths  = 0;
  w->hr_zone[2][0].seconds = BCD(buf[55]);  /* Below zone 3, seconds */
  w->hr_zone[2][0].minutes = BCD(buf[56]);  /* Below zone 3, minutes */
  w->hr_zone[2][0].hours   = BCD(buf[57]);  /* Below zone 3, hours   */
  w->hr_zone[2][1].seconds = BCD(buf[58]);  /* Within zone 3, seconds */
  w->hr_zone[2][1].minutes = BCD(buf[59]);  /* Within zone 3, minutes */
  w->hr_zone[2][1].hours   = BCD(buf[60]);  /* Within zone 3, hours   */
  w->hr_zone[2][2].seconds = BCD(buf[61]);  /* Above zone 3, seconds */
  w->hr_zone[2][2].minutes = BCD(buf[62]);  /* Above zone 3, minutes */
  w->hr_zone[2][2].hours   = BCD(buf[63]);  /* Above zone 3, hours   */

  /* Best lap */

  w->bestlap_split.tenths     = UNIB(buf[65]);
  w->bestlap_split.seconds    = BCD(buf[66]);
  w->bestlap_split.minutes    = BCD(buf[67]);
  w->bestlap_split.hours      = BCD(buf[68]);

  /* energy, total energy (if OwnCal enabled) */

  w->energy       = BCD(buf[69])/10 + 10 * BCD(buf[70]) + 1000 * BCD(buf[71]);
  w->total_energy = BCD(buf[72]) + 100 * BCD(buf[73]) + 10000 * BCD(buf[74]);

  /* cumulative counters */

  w->cumulative_exercise.tenths  = 0;
  w->cumulative_exercise.seconds = 0;
  w->cumulative_exercise.minutes = BCD(buf[77]);
  w->cumulative_exercise.hours   = BCD(buf[75]) + BCD(buf[76]) * 100;

  /* lap data */

  recsiz = 6;

  prev_lap_ascent = prev_lap_dist = 0;
  w->lap_data = calloc(w->laps,sizeof(lap_data_t));

  for ( i = 0; i < w->laps; i++ ) {

    /* set offset */

    offset = S710_HEADER_SIZE_S610 + i * recsiz;
    l = &w->lap_data[i];

    /* timestamp (split) */

    l->cumulative.hours   = buf[offset+2];
    l->cumulative.minutes = buf[offset+1] & 0x3f;
    l->cumulative.seconds = buf[offset] & 0x3f;
    l->cumulative.tenths  = ((buf[offset+1] & 0xc0)>>4) | 
      ((buf[offset] & 0xc0)>>6);

    if ( i == 0 ) 
      memcpy(&l->split,&l->cumulative,sizeof(S710_Time));
    else
      diff_s710_time(&w->lap_data[i-1].cumulative,&l->cumulative,&l->split);

    /* heart rate data */

    l->lap_hr = buf[offset+3];
    l->avg_hr = buf[offset+4];
    l->max_hr = buf[offset+5];
    offset += 6;
  }

  /* sample data */

  offset = S710_HEADER_SIZE_S610 + w->laps * recsiz;

  /* sample size */

  recsiz = 1;

  /* number of samples */

  w->samples = (w->bytes - offset)/recsiz;

  /* allocate memory - bail out if not ok */

  ok = 1;

  MAKEBUF(hr_data,S710_Heart_Rate);

  /* never let a partially allocated workout get through. */

  if ( !ok ) {
    free_workout(w);
    return NULL;
  }

  /* actual sample data - it's in reverse order! */

  for ( i = 0; i < w->samples; i++ ) {
    s = offset + i * recsiz;
    x = w->samples - 1 - i;
    w->hr_data[x] = buf[s];
    s++;
  }

  /* if filtering was requested, filter the HR data and recompute
     avg/max HR values. */

  if ( filter == S710_FILTER_ON ) filter_workout(w);

  return w;
}


/* Write the workout header of an S610 workout in SRD format. */

int
write_workout_header_s610 ( workout_t *w, char *buf )
{
  int tmp1;
  int tmp2;
  int tmp3;

  /* number of bytes in the buffer (including these two) */

  buf[0] = w->bytes & 0xff;
  buf[1] = (w->bytes >> 8) & 0xff;

  /* exercise number and label */

  buf[2] = w->exercise_number;
  if ( w->exercise_number > 0 && w->exercise_number <= 5 ) {
    encode_label(w->exercise_label,&buf[3],7);
  }
  
  /* date of workout */

  buf[10] = HEX(w->date.tm_sec);
  buf[11] = HEX(w->date.tm_min);

  tmp1 = w->date.tm_hour;
  if ( w->ampm & S710_AM_PM_MODE_SET ) {
    if ( w->ampm & S710_AM_PM_MODE_PM ) {
      if ( tmp1 > 12 ) tmp1 -= 12;
    } else {
      if ( tmp1 == 0 ) tmp1 = 12;
    }
  }
  
  buf[12] = HEX(tmp1);
  buf[13] = HEX(w->date.tm_mday) & 0x7f;

  if ( w->ampm & S710_AM_PM_MODE_PM  ) buf[12] |= 0x80;
  if ( w->ampm & S710_AM_PM_MODE_SET ) buf[13] |= 0x80;

  buf[14] = HEX(w->date.tm_year - 100);
  buf[15] = (w->date.tm_mon + 1) & 0x0f;

  /* duration of workout */

  buf[15] |= (w->duration.tenths << 4);
  buf[16] = HEX(w->duration.seconds);
  buf[17] = HEX(w->duration.minutes);
  buf[18] = HEX(w->duration.hours);
  
  buf[19] = w->avg_hr;
  buf[20] = w->max_hr;
  buf[21] = HEX(w->laps);
  buf[22] = HEX(w->laps);

  /* buf[23] is not yet understood */

  buf[24] = HEX(w->user_id);
  if ( w->units.system == S710_UNITS_ENGLISH ) {
    buf[25] |= 0x02;
  }

  switch ( w->recording_interval ) {
  case 60: buf[26] = 2; break;
  case 15: buf[26] = 1; break;
  case  5:
  default: buf[26] = 0; break;
  }

  /* HR limits */

  buf[28] = w->hr_limit[0].lower;
  buf[29] = w->hr_limit[0].upper;
  buf[30] = w->hr_limit[1].lower;
  buf[31] = w->hr_limit[1].upper;
  buf[32] = w->hr_limit[2].lower;
  buf[33] = w->hr_limit[2].upper;

  /* identifies this as an S610 workout (???) */

  buf[34] = 0;
  buf[36] = 251;

  /* time below, within, above hr limits 1 */
  
  buf[37] = HEX(w->hr_zone[0][0].seconds);
  buf[38] = HEX(w->hr_zone[0][0].minutes);
  buf[39] = HEX(w->hr_zone[0][0].hours);
  buf[40] = HEX(w->hr_zone[0][1].seconds);
  buf[41] = HEX(w->hr_zone[0][1].minutes);
  buf[42] = HEX(w->hr_zone[0][1].hours);
  buf[43] = HEX(w->hr_zone[0][2].seconds);
  buf[44] = HEX(w->hr_zone[0][2].minutes);
  buf[45] = HEX(w->hr_zone[0][2].hours);

  /* time below, within, above hr limits 2 */

  buf[46] = HEX(w->hr_zone[1][0].seconds);
  buf[47] = HEX(w->hr_zone[1][0].minutes);
  buf[48] = HEX(w->hr_zone[1][0].hours);
  buf[49] = HEX(w->hr_zone[1][1].seconds);
  buf[50] = HEX(w->hr_zone[1][1].minutes);
  buf[51] = HEX(w->hr_zone[1][1].hours);
  buf[52] = HEX(w->hr_zone[1][2].seconds);
  buf[53] = HEX(w->hr_zone[1][2].minutes);
  buf[54] = HEX(w->hr_zone[1][2].hours);

  /* time below, within, above hr limits 3 */

  buf[55] = HEX(w->hr_zone[2][0].seconds);
  buf[56] = HEX(w->hr_zone[2][0].minutes);
  buf[57] = HEX(w->hr_zone[2][0].hours);
  buf[58] = HEX(w->hr_zone[2][1].seconds);
  buf[59] = HEX(w->hr_zone[2][1].minutes);
  buf[60] = HEX(w->hr_zone[2][1].hours);
  buf[61] = HEX(w->hr_zone[2][2].seconds);
  buf[62] = HEX(w->hr_zone[2][2].minutes);
  buf[63] = HEX(w->hr_zone[2][2].hours);

  /* Best lap */

  buf[64] = (w->bestlap_split.tenths & 0x0f) << 4;
  buf[65] = HEX(w->bestlap_split.seconds);
  buf[66] = HEX(w->bestlap_split.minutes);
  buf[67] = HEX(w->bestlap_split.hours);

  /* energy, total energy (if OwnCal enabled) */

  tmp1 = (w->energy - w->energy % 1000)/1000;
  tmp2 = (w->energy % 1000 - w->energy % 10)/10;
  tmp3 = w->energy % 10;

  buf[69] = HEX(tmp3) << 4;
  buf[70] = HEX(tmp2);
  buf[71] = HEX(tmp1);

  tmp1 = (w->total_energy - w->total_energy % 10000)/10000;
  tmp2 = (w->total_energy % 10000 - w->total_energy % 100)/100;
  tmp3 = w->total_energy % 100;

  buf[72] = HEX(tmp3);
  buf[73] = HEX(tmp2);
  buf[74] = HEX(tmp1);

  /* cumulative counters */

  tmp1 = (w->cumulative_exercise.hours-w->cumulative_exercise.hours % 100)/100;
  tmp2 = w->cumulative_exercise.hours % 100;

  buf[75] = HEX(tmp2);
  buf[76] = HEX(tmp1);
  buf[77] = HEX(w->cumulative_exercise.minutes);

  return S710_HEADER_SIZE_S610;
}


/* 
   "what" is the bitwise or of at least one of:

   S710_WORKOUT_HEADER
   S710_WORKOUT_LAPS
   S710_WORKOUT_SAMPLES

   or it can just be S710_WORKOUT_FULL (everything)
*/

void
print_workout_s610 ( workout_t *w, FILE *fp, int what )
{
  int          i;
  char         buf[BUFSIZ];
  lap_data_t  *l;
  S710_Time    s;

  if ( what & S710_WORKOUT_HEADER ) {

    /* exercise date */
    
    strftime(buf,sizeof(buf),"%Y-%m-%d %H:%M:%S (%a, %d %b %Y)",
             localtime(&w->unixtime));
    fprintf(fp,"Workout date:            %s\n",buf);

    /* HRM type */

    fprintf(fp,"HRM Type:                S610\n");

    /* user id */
    
    fprintf(fp,"User ID:                 %d\n",w->user_id);

    /* exercise number and label */

    if ( w->exercise_number > 0 && w->exercise_number <= 5 ) {
      fprintf(fp,"Exercise:                %d (%s)\n",
	      w->exercise_number,
	      w->exercise_label);
    }

    /* workout mode */
    
    fprintf(fp,"Mode:                    HR");
    fprintf(fp,"\n");
    
    /* exercise duration */
    
    fprintf(fp,"Exercise duration:       ");
    print_s710_time(&w->duration,"hmst",fp);
    fprintf(fp,"\n");

    /* recording interval */
    
    fprintf(fp,"Recording interval:      %d seconds\n",
	    w->recording_interval);
    
    /* average, maximum heart rate */
    
    fprintf(fp,"Average heart rate:      %3d bpm\n",w->avg_hr);
    fprintf(fp,"Maximum heart rate:      %3d bpm\n",w->max_hr);
    
    /* HR limits */
    
    for ( i = 0; i < 3; i++ ) {
      fprintf(fp,"HR Limit %d:              %d to %3d\n",
	      i+1,w->hr_limit[i].lower,w->hr_limit[i].upper);
      fprintf(fp,"\tTime below:      ");
      print_s710_time(&w->hr_zone[i][0],"hms",fp);
      fprintf(fp,"\n");
      fprintf(fp,"\tTime within:     ");
      print_s710_time(&w->hr_zone[i][1],"hms",fp);
      fprintf(fp,"\n");
      fprintf(fp,"\tTime above:      ");
      print_s710_time(&w->hr_zone[i][2],"hms",fp);
      fprintf(fp,"\n");
    }

    /* energy, total energy (units??) */

    fprintf(fp,"Energy:                  %d\n",w->energy);
    fprintf(fp,"Total energy:            %d\n",w->total_energy);
    
    /* cumulative counters */
    
    fprintf(fp,"Cumulative exercise:     ");
    print_s710_time(&w->cumulative_exercise,"hm",fp);
    fprintf(fp,"\n");
    
    /* laps */
    
    fprintf(fp,"Laps:                    %d\n",w->laps);
    fprintf(fp,"\n\n");
  }

  if ( what & S710_WORKOUT_LAPS ) {

    /* lap data */
    
    for ( i = 0; i < w->laps; i++ ) {
      l = &w->lap_data[i];
      fprintf(fp,"Lap %d:\n",i+1);
      fprintf(fp,"\tLap split:          ");
      print_s710_time(&l->split,"hmst",fp);
      fprintf(fp,"\n");
      fprintf(fp,"\tLap cumulative:     ");
      print_s710_time(&l->cumulative,"hmst",fp);
      fprintf(fp,"\n");
      fprintf(fp,"\tLap HR:             %d bpm\n",l->lap_hr);
      fprintf(fp,"\tAverage HR:         %d bpm\n",l->avg_hr);
      fprintf(fp,"\tMaximum HR:         %d bpm\n",l->max_hr);
      fprintf(fp,"\n");
    }
  }

  if ( what & S710_WORKOUT_SAMPLES ) {

    /* sample data */
    
    fprintf(fp,"\nRecorded data:\n\n");
    fprintf(fp,"Time\t\t HR");
    fprintf(fp,"\n");

    memset(&s,0,sizeof(s));
    for ( i = 0; i < w->samples; i++ ) {
      print_s710_time(&s,"hms",fp);
      fprintf(fp,"\t\t%3d",w->hr_data[i]);
      fprintf(fp,"\n");

      increment_s710_time(&s,w->recording_interval);
    }
  }
}


/* 
   This function takes a workout_t *w and dumps it to FILE *fp in HRM format.
   
   For a description of the HRM file format, see:

   http://developer.polar.fi/developer.nsf/main

   The HRM file format version produced is version 1.06.

   Note: According to the HRM file format spec, we need CR/LF ("\r\n")
   to terminate lines.  So that's why there are all those \r's floating
   around in there, in case you were wondering.
*/

void
print_workout_as_hrm_s610 ( workout_t *w, FILE *fp )
{
  int i, j;
  int
    above_max,
    upper_to_max,
    lower_to_upper,
    rest_to_lower,
    below_rest;
  int sampled_seconds;
  int lap_start_sample;
  int lap_end_sample;
  int lap_min_hr;

  /* sanity checks. */

  if ( w == NULL || fp == NULL ) {
    fprintf(stderr,"print_hrm_as_s710: improper usage(%p,%p)\n",w,fp);
    return;
  }

  if ( w->type != S710_HRM_S610 ) {
    fprintf(stderr,"print_hrm_as_s610: input is not from S610/S610i\n");
    return;
  }

  /* Enough goofing off, let's get on with it. */

  /* Print version and monitor numbers */

  fprintf(fp,"[Params]\r\nVersion=106\r\nMonitor=%d\r\n",w->type);

  /* Print the SMode flags (to our knowledge) */

  fprintf(fp,"SMode=0000000%d0\r\n",
	  (w->units.system == S710_UNITS_ENGLISH) ? 1 : 0);
  
  /* Date, StartTime, Length, Interval */

  fprintf(fp,"Date=%4d%02d%02d\r\n",
	  w->date.tm_year + 1900,
	  w->date.tm_mon + 1,
	  w->date.tm_mday);

  fprintf(fp,"StartTime=%02d:%02d:%02d.0\r\n",
	  w->date.tm_hour,
	  w->date.tm_min,
	  w->date.tm_sec);

  fprintf(fp,"Length=%02d:%02d:%02d.%d\r\n",
	  w->duration.hours,
	  w->duration.minutes,
	  w->duration.seconds,
	  w->duration.tenths);

  fprintf(fp,"Interval=%d\r\n",w->recording_interval);
	  
  /* Limits 1, 2 and 3 */

  fprintf(fp,
	  "Upper1=%d\r\nLower1=%d\r\n"
	  "Upper2=%d\r\nLower2=%d\r\n"
	  "Upper3=%d\r\nLower3=%d\r\n",
	  w->hr_limit[0].upper,
	  w->hr_limit[0].lower,
	  w->hr_limit[1].upper,
	  w->hr_limit[1].lower,
	  w->hr_limit[2].upper,
	  w->hr_limit[2].lower);

  /* Exercise timers 1, 2 and 3 */

  fprintf(fp,
	  "Timer1=00:00:00.0\r\n"
	  "Timer2=00:00:00.0\r\n"
	  "Timer3=00:00:00.0\r\n");

  /* ActiveLimit - I think this is set using the Polar software */

  fprintf(fp,"ActiveLimit=0\r\n");

  /* 
     MaxHR, RestHR are not recorded in the file.  They're recorded in
     the user settings, but instead of downloading them from the watch,
     we just put default values here.  Someday we may revisit this and
     set up a way for users to have a .s710rc configuration file or
     something like that.  That way, this sort of thing can be read
     from the user's config file.
  */

  fprintf(fp,"MaxHR=%d\r\n",S710_HRM_MAX_HR);
  fprintf(fp,"RestHR=%d\r\n",S710_HRM_REST_HR);
  fprintf(fp,"StartDelay=0\r\n");

  /* VO2Max, Weight - see above. */

  fprintf(fp,"VO2max=50\r\n");
  fprintf(fp,"Weight=%d\r\n",
	  (w->units.system == S710_UNITS_ENGLISH) ? 150 : 70);

  /* Note */

  fprintf(fp,"\r\n[Note]\r\n");

  /* IntTimes */

  fprintf(fp,"\r\n[IntTimes]\r\n");
  
  for ( i = 0; i < w->laps; i++ ) {

    /* 
       This is really annoying, but we actually have to compute the 
       value of the minimum heart rate for each lap.  We can only do
       this if we know the recording interval (which we should always
       know).
    */

    lap_min_hr = 0;

    if ( w->recording_interval != 0 ) {
      lap_start_sample = w->lap_data[i].split.seconds +
	60 * (w->lap_data[i].split.minutes + 
	      60 * w->lap_data[i].split.hours);
      lap_end_sample = w->lap_data[i].cumulative.seconds +
	60 * (w->lap_data[i].cumulative.minutes + 
	      60 * w->lap_data[i].cumulative.hours);

      lap_start_sample /= w->recording_interval;
      lap_end_sample /= w->recording_interval;

      lap_start_sample = lap_end_sample - lap_start_sample;

      if ( lap_end_sample >= w->samples ) lap_end_sample = w->samples - 1;
      if ( lap_start_sample < 0 ) lap_start_sample = 0;

      lap_min_hr = 255;
      for ( j = lap_start_sample; j <= lap_end_sample; j++ ) {
	if ( w->hr_data[j] != 0 && w->hr_data[j] < lap_min_hr ) {
	  lap_min_hr = w->hr_data[j];
	}
      }
    }

    fprintf(fp,"%02d:%02d:%02d.%d\t%d\t%d\t%d\t%d\r\n",
	    w->lap_data[i].cumulative.hours,
	    w->lap_data[i].cumulative.minutes,
	    w->lap_data[i].cumulative.seconds,
	    w->lap_data[i].cumulative.tenths,
	    w->lap_data[i].lap_hr,
	    lap_min_hr,
	    w->lap_data[i].avg_hr,
	    w->lap_data[i].max_hr);
    fprintf(fp,"2\t0\t0\t0\t0\t0\r\n");
    fprintf(fp,"0\t0\t0\t0\t0\r\n");
    fprintf(fp,"0\t0\t0\t0\t0\t0\r\n");
    fprintf(fp,"0\t0\t0\t0\t0\t0\r\n");
  }

  /* IntNotes */

  fprintf(fp,"\r\n[IntNotes]\r\n");

  /* ExtraData */

  fprintf(fp,"\r\n[ExtraData]\r\n");

  /* Summary-123 */

  fprintf(fp,"\r\n[Summary-123]\r\n");

  /* We'll need this a few times. */

  sampled_seconds = (w->samples - 1) * w->recording_interval;

  for ( i = 0; i < 3; i++ ) {
    
    /* 
       We actually have to compute these values.  Actually, one of them
       shouldn't need to be computed, but while we're at it, why not. 
    */

    above_max      = 0;
    upper_to_max   = 0;
    lower_to_upper = 0;
    rest_to_lower  = 0;
    below_rest     = 0;

    /* We'll need this a few times. */
    
    /* 
       We start at sample 1.  Each sample covers the preceding five
       seconds.
    */

    for ( j = 1; j < w->samples; j++ ) {
      if      ( w->hr_data[j] > S710_HRM_MAX_HR )      above_max++;
      else if ( w->hr_data[j] > w->hr_limit[i].upper ) upper_to_max++;
      else if ( w->hr_data[j] > w->hr_limit[i].lower ) lower_to_upper++;
      else if ( w->hr_data[j] > S710_HRM_REST_HR )     rest_to_lower++;
      else                                             below_rest++;
    }

    /* Now we have the number of samples in each of the five ranges. */

    fprintf(fp,"%d\t%d\t%d\t%d\t%d\t%d\r\n",
	    sampled_seconds,
	    above_max * w->recording_interval,
	    upper_to_max * w->recording_interval,
	    lower_to_upper * w->recording_interval,
	    rest_to_lower * w->recording_interval,
	    below_rest * w->recording_interval);
    fprintf(fp,"%d\t%d\t%d\t%d\r\n",
	    S710_HRM_MAX_HR,
	    w->hr_limit[i].upper,
	    w->hr_limit[i].lower,
	    S710_HRM_REST_HR);
  }
  fprintf(fp,"0\t%d\r\n",w->samples-1);

  /* Summary-TH - see comments below. */

  fprintf(fp,"\r\n[Summary-TH]\r\n");
  fprintf(fp,"%d\t0\t%d\t0\t0\t0\r\n",sampled_seconds,sampled_seconds);
  fprintf(fp,"%d\t0\t0\t%d\r\n",S710_HRM_MAX_HR,S710_HRM_REST_HR);
  fprintf(fp,"0\t%d\r\n",w->samples-1);

  /* 
     HRZones - these quantities are set by the user in the Polar software.
     Fill in with an "empty" set of values.
  */

  fprintf(fp,"\r\n[HRZones]\r\n");
  fprintf(fp,"0\r\n0\r\n0\r\n0\r\n0\r\n0\r\n0\r\n0\r\n0\r\n0\r\n0\r\n");

  /* HRData */

  fprintf(fp,"\r\n[HRData]\r\n");

  for ( i = 0; i < w->samples; i++ ) {
    fprintf(fp,"%d\r\n",w->hr_data[i]);
  }

  /* That's all, folks. */

  fflush(fp);
}

