/* $Id$ */

#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "s710.h"
#include "config.h"


#ifndef S710_HRM_MAX_HR
#define S710_HRM_MAX_HR  200
#endif

#ifndef S710_HRM_REST_HR
#define S710_HRM_REST_HR  60
#endif


#define  MAKEBUF(a,b) \
if ( (w->a = calloc(w->samples,sizeof(b))) == NULL ) { \
  fprintf(stderr,"%s: calloc(%d,%ld): %s\n", \
          #a,w->samples,(long)sizeof(b),strerror(errno)); \
  ok = 0; \
}
                     

/* don't call this unless you're sure it's going to be ok. */

workout_t *
extract_workout_s625x ( unsigned char *buf, S710_Filter filter )
{
  int              i;
  int              offset;
  int              s;
  int              x;
  int              recsiz;
  workout_t       *w;
  lap_data_t      *l;
  S710_Distance    prev_lap_dist;
  S710_Altitude    prev_lap_ascent;
  unsigned long    accum;
  int              ok;
  int              j;
  float            avg;
  int              ax;
  int              cadence_en, power_en, speed_en, altitude_en;

  if ( (w = calloc(1,sizeof(workout_t))) == NULL ) {
    fprintf(stderr,"extract_workout: calloc(%ld): %s\n",
	    (long)sizeof(workout_t),strerror(errno));
    return NULL;
  }

  /* define the type of the HRM */

  w->type = S710_HRM_S625X;

  /* number of bytes in the buffer (including these two) */

  w->bytes = buf[0] + (buf[1]<<8);

  w->exercise_number = buf[2];
  if ( w->exercise_number > 0 && w->exercise_number <= 5 ) {
    extract_label(&buf[3],&w->exercise_label,7);
  } else {
    strcpy(w->exercise_label,"<empty>");
  }
  
  /* date of workout */

  w->date.tm_sec   = BCD(buf[10]);
  w->date.tm_min   = BCD(buf[11]);
  w->date.tm_hour  = BCD(buf[12] & 0x7f);

  /* PATCH for AM/PM mode detection from Berend Ozceri */
  
  w->ampm = S710_AM_PM_MODE_UNSET;
  if ( buf[13] & 0x80 ) w->ampm |= S710_AM_PM_MODE_SET;
  if ( buf[12] & 0x80 ) w->ampm |= S710_AM_PM_MODE_PM;

  w->date.tm_hour += (buf[13] & 0x80) ?                     /* am/pm mode?   */
    ((buf[12] & 0x80) ? ((w->date.tm_hour < 12) ? 12 : 0) : /* yes, pm set   */
     ((w->date.tm_hour >= 12) ? -12 : 0)) :                 /* yes, pm unset */
    0;                                                      /* no            */

  w->date.tm_mon   = LNIB(buf[15]) - 1;
  w->date.tm_mday  = BCD(buf[13] & 0x7f);
  w->date.tm_year  = 100 + BCD(buf[14]);
  w->date.tm_isdst = -1; /* Daylight savings time not known yet? */
  w->unixtime = mktime(&w->date);

  /* duration of workout */

  w->duration.tenths       = UNIB(buf[15]);
  w->duration.seconds      = BCD(buf[16]);
  w->duration.minutes      = BCD(buf[17]);
  w->duration.hours        = BCD(buf[18]);

  /* avg, max HR */
  
  w->avg_hr                = buf[19];
  w->max_hr                = buf[20];

  /* number of laps */

  w->laps                  = BCD(buf[21]);

  /* user ID */

  w->user_id               = BCD(buf[24]);

  /* units */

  if ( buf[25] & 0x02 ) { /* english */

    w->units.system      = S710_UNITS_ENGLISH;
    w->units.altitude    = S710_ALTITUDE_FT;
    w->units.speed       = S710_SPEED_MPH;
    w->units.distance    = S710_DISTANCE_MI;
    w->units.temperature = S710_TEMPERATURE_F;
    ax                   = 5;

  } else {         /* metric */

    w->units.system      = S710_UNITS_METRIC;
    w->units.altitude    = S710_ALTITUDE_M;
    w->units.speed       = S710_SPEED_KPH;
    w->units.distance    = S710_DISTANCE_KM;
    w->units.temperature = S710_TEMPERATURE_C;
    ax                   = 1;
  }

  /* recording mode and interval */

  w->mode     = buf[26];
  cadence_en  = w->mode & S710_MODE_CADENCE;
  power_en    = w->mode & S710_MODE_POWER;
  speed_en    = w->mode & S710_MODE_SPEED;
  altitude_en = w->mode & S710_MODE_ALTITUDE;
  
  switch ( buf[27] ) {
  case 0:  w->recording_interval = 5;  break;
  case 1:  w->recording_interval = 15; break;
  case 2:  w->recording_interval = 60; break;
  default: w->recording_interval = 0;  break;
  }

  /* HR limits */

  w->hr_limit[0].lower     = buf[29];
  w->hr_limit[0].upper     = buf[30];
  w->hr_limit[1].lower     = buf[31];
  w->hr_limit[1].upper     = buf[32];
  w->hr_limit[2].lower     = buf[33];
  w->hr_limit[2].upper     = buf[34];

  /* time below, within, above hr limits 1 */
  
  w->hr_zone[0][0].tenths  = 0;
  w->hr_zone[0][0].seconds = BCD(buf[38]);  /* Below zone 1, seconds */
  w->hr_zone[0][0].minutes = BCD(buf[39]);  /* Below zone 1, minutes */
  w->hr_zone[0][0].hours   = BCD(buf[40]);  /* Below zone 1, hours   */
  w->hr_zone[0][1].seconds = BCD(buf[41]);  /* Within zone 1, seconds */
  w->hr_zone[0][1].minutes = BCD(buf[42]);  /* Within zone 1, minutes */
  w->hr_zone[0][1].hours   = BCD(buf[43]);  /* Within zone 1, hours   */
  w->hr_zone[0][2].seconds = BCD(buf[44]);  /* Above zone 1, seconds */
  w->hr_zone[0][2].minutes = BCD(buf[45]);  /* Above zone 1, minutes */
  w->hr_zone[0][2].hours   = BCD(buf[46]);  /* Above zone 1, hours   */

  /* time below, within, above hr limits 2 */

  w->hr_zone[1][0].tenths  = 0;
  w->hr_zone[1][0].seconds = BCD(buf[47]);  /* Below zone 2, seconds */
  w->hr_zone[1][0].minutes = BCD(buf[48]);  /* Below zone 2, minutes */
  w->hr_zone[1][0].hours   = BCD(buf[49]);  /* Below zone 2, hours   */
  w->hr_zone[1][1].seconds = BCD(buf[50]);  /* Within zone 2, seconds */
  w->hr_zone[1][1].minutes = BCD(buf[51]);  /* Within zone 2, minutes */
  w->hr_zone[1][1].hours   = BCD(buf[52]);  /* Within zone 2, hours   */
  w->hr_zone[1][2].seconds = BCD(buf[53]);  /* Above zone 2, seconds */
  w->hr_zone[1][2].minutes = BCD(buf[54]);  /* Above zone 2, minutes */
  w->hr_zone[1][2].hours   = BCD(buf[55]);  /* Above zone 2, hours   */

  /* time below, within, above hr limits 3 */

  w->hr_zone[2][0].tenths  = 0;
  w->hr_zone[2][0].seconds = BCD(buf[56]);  /* Below zone 3, seconds */
  w->hr_zone[2][0].minutes = BCD(buf[57]);  /* Below zone 3, minutes */
  w->hr_zone[2][0].hours   = BCD(buf[58]);  /* Below zone 3, hours   */
  w->hr_zone[2][1].seconds = BCD(buf[59]);  /* Within zone 3, seconds */
  w->hr_zone[2][1].minutes = BCD(buf[60]);  /* Within zone 3, minutes */
  w->hr_zone[2][1].hours   = BCD(buf[61]);  /* Within zone 3, hours   */
  w->hr_zone[2][2].seconds = BCD(buf[62]);  /* Above zone 3, seconds */
  w->hr_zone[2][2].minutes = BCD(buf[63]);  /* Above zone 3, minutes */
  w->hr_zone[2][2].hours   = BCD(buf[64]);  /* Above zone 3, hours   */

  /* Best lap */

  w->bestlap_split.tenths     = UNIB(buf[66]);
  w->bestlap_split.seconds    = BCD(buf[67]);
  w->bestlap_split.minutes    = BCD(buf[68]);
  w->bestlap_split.hours      = BCD(buf[69]);

  /* energy, total energy (if OwnCal enabled) */

  w->energy       = BCD(buf[70])/10 + 10 * BCD(buf[71]) + 1000 * BCD(buf[72]);
  w->total_energy = BCD(buf[73]) + 100 * BCD(buf[74]) + 10000 * BCD(buf[75]);

  /* cumulative counters */

  w->cumulative_exercise.tenths  = 0;
  w->cumulative_exercise.seconds = 0;
  w->cumulative_exercise.minutes = BCD(buf[78]);
  w->cumulative_exercise.hours   = BCD(buf[76]) + BCD(buf[77]) * 100;

  w->cumulative_ride.tenths      = 0;
  w->cumulative_ride.seconds     = 0;
  w->cumulative_ride.minutes     = BCD(buf[81]);
  w->cumulative_ride.hours       = BCD(buf[79]) + BCD(buf[80]) * 100;

  w->odometer = 10000 * BCD(buf[84]) + 100 * BCD(buf[83]) + BCD(buf[82]);

  /* exercise distance */

  w->exercise_distance = buf[85] + (buf[86]<<8);

  /* avg and max speed */

  w->avg_speed = buf[87] | ((buf[88] & 0xfUL) << 8);
  w->max_speed = (buf[89] << 4) | (buf[88] >> 4);

  /* avg, max cadence */

  w->avg_cad  = buf[90];
  w->max_cad  = buf[91];

  /* min, avg, max temperature */

  if ( w->units.system == S710_UNITS_ENGLISH ) {

    /* English units */

    w->min_temp = buf[98];
    w->avg_temp = buf[99];
    w->max_temp = buf[100];

  } else {

    /* Metric units */

    w->min_temp = buf[98] & 0x7f;
    w->min_temp = ( buf[98] & 0x80 ) ? w->min_temp : - w->min_temp;
    
    w->avg_temp = buf[99] & 0x7f;
    w->avg_temp = ( buf[99] & 0x80 ) ? w->avg_temp : - w->avg_temp;
    
    w->max_temp = buf[100] & 0x7f;
    w->max_temp = ( buf[100] & 0x80 ) ? w->max_temp : - w->max_temp;
  }

  /* altitude, ascent */

  w->min_alt  = buf[92] + ((buf[93] & 0x7f)<<8);
  w->min_alt  = ( buf[93] & 0x80 ) ? w->min_alt : - w->min_alt;

  w->avg_alt  = buf[94] + ((buf[95] & 0x7f)<<8);
  w->avg_alt  = ( buf[95] & 0x80 ) ? w->avg_alt : - w->avg_alt;

  w->max_alt  = buf[96] + ((buf[97] & 0x7f)<<8);
  w->max_alt  = ( buf[97] & 0x80 ) ? w->max_alt : - w->max_alt;
  
  w->ascent   = (buf[102] << 8) + buf[101];

  /* avg, max power data */

  w->avg_power.power       = buf[103] + (LNIB(buf[104]) << 8);
  w->max_power.power       = UNIB(buf[104]) + (buf[105] << 4);
  w->avg_power.pedal_index = buf[106];
  w->max_power.pedal_index = buf[107];
  w->avg_power.lr_balance  = buf[108];
  /* there is no max_power LR balance */
  w->max_power.lr_balance  = 0;

  /* lap data */

  recsiz = 6;
  if (altitude_en) recsiz += 5;
  if (speed_en) {
    if (cadence_en) recsiz += 1;
    if (power_en) recsiz += 4;
    recsiz += 4;
  }

  prev_lap_ascent = prev_lap_dist = 0;
  w->lap_data = calloc(w->laps,sizeof(lap_data_t));

  for ( i = 0; i < w->laps; i++ ) {

    /* position to the start of the lap */

    offset = S710_HEADER_SIZE_S625X + i * recsiz;
    l = &w->lap_data[i];

    /* timestamp (split) */

    l->cumulative.hours   = buf[offset+2];
    l->cumulative.minutes = buf[offset+1] & 0x3f;
    l->cumulative.seconds = buf[offset] & 0x3f;
    l->cumulative.tenths  = ((buf[offset+1] & 0xc0)>>4) | 
      ((buf[offset] & 0xc0)>>6);

    if ( i == 0 ) 
      memcpy(&l->split,&l->cumulative,sizeof(S710_Time));
    else
      diff_s710_time(&w->lap_data[i-1].cumulative,&l->cumulative,&l->split);

    /* heart rate data */

    l->lap_hr = buf[offset+3];
    l->avg_hr = buf[offset+4];
    l->max_hr = buf[offset+5];
    offset += 6;

    /* altitude data */

    if (altitude_en) {
      l->alt = buf[offset] + (buf[offset+1]<<8) - 512;
      /* This ascent data is cumulative from start of the exercise */
      l->cumul_ascent = buf[offset+2] + (buf[offset+3]<<8);
      l->ascent = l->cumul_ascent - prev_lap_ascent;
      prev_lap_ascent = l->cumul_ascent;

      if ( w->units.system == S710_UNITS_ENGLISH ) {  /* English units */
	l->temp = buf[offset+4] + 14;
	l->alt *= ax;
      } else {
	l->temp = buf[offset+4] - 10;
      }

      offset += 5;
    }

    /* bike data */

    if (speed_en) {

      /* cadence data */

      if (cadence_en) { l->cad = buf[offset]; offset += 1; }

      /* next 4 bytes are power data */

      if (power_en) { 
	l->power.power = buf[offset] + (buf[offset+1]<<8);
	l->power.lr_balance = buf[offset+3];   /* ??? switched  ??? */
	l->power.pedal_index = buf[offset+2];  /* ??? with this ??? */
	offset += 4;
      }

      /* 
         next 4 bytes are distance/speed data
         this is cumulative distance from start (at least with S720i)
      */
      l->cumul_distance = buf[offset] + (buf[offset+1]<<8);
      l->distance = l->cumul_distance - prev_lap_dist;
      prev_lap_dist = l->cumul_distance;

      /* 
	 The offset + 2 appears to be 4b.4b, where the upper 4b is the
	 integer portion of the speed and the lower 4b is the fractional
	 portion in sixteenths.  I'm not sure where the most significant
	 bits come from, but it might be the high order nibble of the next
	 byte- lets try that.
      */
      
      l->speed = buf[offset+2] + ((buf[offset+3] & 0xf0) << 4);

      /* fprintf(stderr, "distance/speed: %02x %02x %02x %02x\n",
	 buf[offset], buf[offset+1], buf[offset+2], buf[offset+3]); */
      /*  fprintf(stderr, "SPEED: %X\n", l->speed); */
      
    }
  }

  /* sample data */

  offset = S710_HEADER_SIZE_S625X + w->laps * recsiz;

  /* sample size */

  recsiz = 1;
  if (altitude_en) recsiz += 2;
  if (speed_en) {
    if (altitude_en) recsiz -= 1;
    recsiz += 2;
    if (power_en) recsiz += 4;
    if (cadence_en) recsiz += 1;
  }

  /* number of samples */

  w->samples = (w->bytes - offset)/recsiz;

  /* allocate memory - bail out if not ok */

  ok = 1;

  MAKEBUF(hr_data,S710_Heart_Rate);
  if (altitude_en) MAKEBUF(alt_data, S710_Altitude);
  if (speed_en) {
    MAKEBUF(speed_data, S710_Speed);
    MAKEBUF(dist_data, S710_Distance);
    if (power_en)   MAKEBUF(power_data, S710_Power);
    if (cadence_en) MAKEBUF(cad_data, S710_Cadence);
  }

  /* never let a partially allocated workout get through. */

  if ( !ok ) {
    free_workout(w);
    return NULL;
  }

  /* actual sample data - it's in reverse order! */

  for ( i = 0; i < w->samples; i++ ) {
    s = offset + i * recsiz;
    x = w->samples - 1 - i;
    w->hr_data[x] = buf[s];
    s++;

    if (altitude_en) {
      w->alt_data[x] = buf[s] + ((buf[s+1] & 0x1f)<<8) - 512;
      w->alt_data[x] *= ax;
      s += 2;
    }

    if (speed_en) {
      if (altitude_en) s -= 1;
      w->speed_data[x] = ((buf[s] & 0xe0) << 3) + buf[s+1];
      s += 2;
      if (power_en) { 
	w->power_data[x].power = buf[s] + (buf[s+1]<<8);
	w->power_data[x].lr_balance = buf[s+2];
	w->power_data[x].pedal_index = buf[s+3];
	s += 4;
      }
      if (cadence_en) w->cad_data[x] = buf[s];
    }
  }

  if (speed_en) {
    accum = 0;
    for ( i = 0; i < w->samples; i++ ) {
      w->dist_data[i] = accum / 57600.0;
      accum += w->speed_data[i] * w->recording_interval;
    }
  }

  /* if filtering was requested, filter the HR data and recompute
     avg/max HR values. */

  if ( filter == S710_FILTER_ON ) filter_workout(w);

  /* compute median speed and highest sampled speed */

  if (speed_en) {
    avg = 0;
    w->highest_speed = 0;
    j = 0;
    for ( i = 0; i < w->samples; i++ ) {
      if ( w->speed_data[i] > w->highest_speed ) 
	w->highest_speed = w->speed_data[i];
      if ( w->speed_data[i] > 0 ) {
	avg = (float)(avg * j + w->speed_data[i])/(j+1);
	j++;
      }
    }
    w->median_speed = avg;
  }

  return w;
}


